// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Forward declarations
@class DGSMonXModel;
@class DGSMonXPreferences;
@class DGSMonXThread;


// -----------------------------------------------------------------------------
/// @brief The DGSMonXController class implements the Controller role of the
/// design pattern commonly known as Model-View-Controller.
///
/// DGSMonXController reacts to user input in the GUI and controls the
/// program flow resulting from this user input.
///
/// DGSMonXController also implements the role of delegate of the
/// NSApplication main object (configured in MainMenu.xib).
///
/// In addition, DGSMonXController implements the following protocols and
/// interfaces, and has the following responsibilities:
/// - @e NSNibAwaking protocol: to do stuff after the class has been loaded and
///   instantiated from the .nib file
/// - @e NSMenuValidation protocol: for all menu items that have this class set
///   as their target, the method validateMenuItem; is queried to automatically
///   enable or disable those menu items
/// - Central handling for severe application errors; this class registers
///   with the default notification center for the notification
///   #errorConditionOccurredNotification. Classes may post such a notification
///   if they detect an error they cannot handle
///
/// @note This class is instantiated when MainMenu.xib is loaded.
// -----------------------------------------------------------------------------
@interface DGSMonXController : NSObject
{
@private
  /// @name Outlets
  /// @brief These variables are outlets and therefore initialized in
  /// MainMenu.xib
  //@{
  DGSMonXModel* m_theModel;
  DGSMonXPreferences* m_thePreferences;
  NSWindow* m_textViewWindow;   ///< @brief Window that contains #m_textView.
  NSTextView* m_textView;       ///< @brief Used to display the content of a text file (e.g. COPYING).
  NSMenu* m_theMenu;
  NSMenuItem* m_checkOrCancelAllMenuItem;
  //@}

  /// @name Other members
  //@{
  NSStatusItem* m_statusItem;
  NSConnection* m_connection; 
  DGSMonXThread* m_monitoringThread;
  int m_monitoringThreadState;
  NSTimer* m_statusItemAnimationTimer;
  enum DGSMonXStatusItemState m_statusItemState;
  NSImage* m_normalStateImageBlackAndWhite;
  NSImage* m_normalStateImageGrayScale;
  NSImage* m_itsYourTurnStateImage;
  //@}
}

/// @name Initializers
//@{
- (id) init;
//@}

/// @name Messages sent by the monitoring thread
//@{
- (void) setMonitoringThread:(id)anObject;
- (NSArray*) serverList;
- (void) setMonitoringThreadState:(int)state;
- (void) gameStateCheckDidFinish;
//@}

@end
