// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXModel.h"
#import "DGSMonXServer.h"


/// @brief This category declares private methods for the DGSMonXModel
/// class. 
@interface DGSMonXModel()
/// @name Initialization and deallocation
//@{
- (void) dealloc;
//@}
@end


@implementation DGSMonXModel

// -----------------------------------------------------------------------------
/// @brief Initializes an DGSMonXModel object.
///
/// @note This is the designated initializer of DGSMonXModel.
// -----------------------------------------------------------------------------
- (id) init
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;

  // Initialize list to nil
  // -> loadServerListFromUserDefaults() will be called at the appropriate time
  m_serverList = nil;

  // Return 
  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXModel object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  // When the array is deallocated, it releases all items for us
  if (m_serverList)
    [m_serverList autorelease];
  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Initializes default values in this model from the user's defaults
/// (i.e. [NSUserDefaults standardUserDefaults]).
// -----------------------------------------------------------------------------
- (void) loadServerListFromUserDefaults
{
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  if (! userDefaults)
    return;

  // Create the array that we are going to fill. We do not need to retain here,
  // this will be done further down when we call setServerList:().
  NSMutableArray* serverList = [NSMutableArray array];

  // Get data from user defaults
  NSArray* userDefaultsServerList = [userDefaults arrayForKey:serverListKey];
  NSEnumerator* enumerator = [userDefaultsServerList objectEnumerator];
  NSDictionary* userDefaultsDict;
  while (userDefaultsDict = (NSDictionary*)[enumerator nextObject])
  {
    DGSMonXServer* server = [[DGSMonXServer alloc] initWithDictionary:userDefaultsDict];
    // We want the array to retain and release the object for us -> decrease
    // the retain count by 1 (was set to 1 by alloc/init)
    [server autorelease];
    [serverList addObject:server];
  }

  // IMPORTANT: Do ***NOT*** assign directly to m_serverList!!! Instead use
  // the set-accessor so that the array controller defined in MainMenu.xib
  // takes notice of the change through the key-value-observing mechanism.
  [self setServerList:serverList];
}

// -----------------------------------------------------------------------------
/// @brief Writes current values in this model to the User Default system's
/// application domain.
///
/// @note A value is written only if the current value is different from the
/// value already stored in the User Defaults system.
// -----------------------------------------------------------------------------
- (void) writeServerListToUserDefaults
{
  // Collect data that is going to be written
  NSMutableArray* userDefaultsServerList = [NSMutableArray array];
  NSArray* serverList = [self serverList];
  NSEnumerator* enumerator = [serverList objectEnumerator];
  DGSMonXServer* server;
  while (server = (DGSMonXServer*)[enumerator nextObject])
    [userDefaultsServerList addObject:[server asDictionary]];

  // Write to user defaults
  // Note: NSUserDefaults takes care entirely by itself of writing only changed
  // values, neither this method here nor DGSMonXServer contain a single line
  // of code supporting this feature. Ain't that cool?! :-)
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  [userDefaults setObject:userDefaultsServerList forKey:serverListKey];
}

// -----------------------------------------------------------------------------
/// @brief Returns the list of DGSMonXServer objects managed by this model.
// -----------------------------------------------------------------------------
- (NSArray*) serverList
{
  return [[m_serverList retain] autorelease];
}

// -----------------------------------------------------------------------------
/// @brief Configures this model with a list of DGSMonXServer objects.
///
/// @note This method is required for this model to be KVC compliant, so that
/// this model can be used with Cocoa bindings.
// -----------------------------------------------------------------------------
- (void) setServerList:(NSArray*)serverList
{
  if (m_serverList == serverList)
    return;
  if (m_serverList)
    [m_serverList autorelease];
  if (serverList)
    m_serverList = (NSMutableArray*)[serverList retain];
  else
    m_serverList = nil;
}

// -----------------------------------------------------------------------------
/// @brief Always returns false.
///
/// This method can be used to permanently disable a GUI element in the nib
/// file. To do it, bind the GUI element's "enabled" binding to the model key
/// path named after this method.
// -----------------------------------------------------------------------------
- (bool) isEnabled
{
  return false;
}

@end
