// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------



// -----------------------------------------------------------------------------
/// @brief The DGSMonXPreferences class is a mini-controller for the user
/// preferences dialog. It is also responsible to set up user defaults in the
/// NSRegistration domain.
///
/// Large parts of the user preferences dialog are managed by the
/// "Cocoa Bindings" mechanism. The controller activities of DGSMonXPreferences
/// are therefore only minimal:
/// - Implement action methods for a few buttons
/// - Provide key-value coding and validation methods for properties that
///   require input validation
/// - Provide key-value coding methods that can be queried by Cocoa Bindings to
///   enable/disable dependent controls
///
/// @note DGSMonXController is set up in MainMenu.xib as the delegate of the
/// preferences dialog (which is a NSWindow).
///
/// @note This class is instantiated when MainMenu.xib is loaded.
// -----------------------------------------------------------------------------
@interface DGSMonXPreferences : NSObject
{
@private
  // These variables are outlets and therefore initialized in MainMenu.xib
  NSWindow* m_preferencesDialog;
  NSArrayController* m_serversController;
  // These variables store property values
  bool statusItemIconStyleEnabled;
  bool statusItemNormalTextEnabled;
  bool statusItemItsYourTurnTextEnabled;
  bool notifyAnimateStatusItemIntervalEnabled;
  int minimumValueMonitoringIntervalAmount;
}

/// @name Initializers
//@{
- (id) init;
//@}
/// @name Open/close dialog
//@{
- (void) showPreferencesDialog;
- (void) preferencesDialogCloseClicked:(id)sender;
- (void) testClicked:(id)sender;
//@}
/// @name Input validation
//@{
- (enum DGSMonXStatusItemStyle) statusItemStyle;
- (void) setStatusItemStyle:(enum DGSMonXStatusItemStyle)statusItemStyle;
- (NSString*) statusItemNormalText;
- (void) setStatusItemNormalText:(NSString*)text;
- (BOOL) validateStatusItemNormalText:(id*)valueObject error:(NSError**)outError;
- (NSString*) statusItemItsYourTurnText;
- (void) setStatusItemItsYourTurnText:(NSString*)text;
- (BOOL) validateStatusItemItsYourTurnText:(id*)valueObject error:(NSError**)outError;
- (enum DGSMonXNotificationStyle) notifyStyle;
- (void) setNotifyStyle:(enum DGSMonXNotificationStyle)notificationStyle;
- (int) monitoringIntervalUnit;
- (void) setMonitoringIntervalUnit:(int)monitoringIntervalUnit;
//@}
/// @name Control enablers
//@{
@property(nonatomic, assign) bool statusItemIconStyleEnabled;
@property(nonatomic, assign) bool statusItemNormalTextEnabled;
@property(nonatomic, assign) bool statusItemItsYourTurnTextEnabled;
@property(nonatomic, assign) bool notifyAnimateStatusItemIntervalEnabled;
//@}
/// @name Other properties
//@{
@property(nonatomic, assign) int minimumValueMonitoringIntervalAmount;
//@}
/// @name Other methods
//@{
+ (void) setupDefaults;
//@}

@end
