// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXPreferences.h"
#import "DGSMonXServer.h"
#import "DGSMonXUserDefaultsUpdater.h"


/// @brief This category declares private methods for the DGSMonXPreferences
/// class. 
@interface DGSMonXPreferences()
/// @name Initialization and deallocation
//@{
- (void) dealloc;
//@}
/// @name NSWindow delegate
//@{
- (void) windowWillClose:(NSNotification*)notification;
//@}
/// @name Private helpers
//@{
- (BOOL) validateStatusItemText:(id*)valueObject error:(NSError**)outError;
- (void) updateEnablersDependingOnStatusItemStyle;
- (void) updateEnablersDependingOnNotifyStyle;
- (void) updateMinimumValueMonitoringIntervalAmount;
- (void) adjustMonitoringIntervalAmountIfExceedsMinimumValue;
//@}
@end


@implementation DGSMonXPreferences

@synthesize statusItemIconStyleEnabled;
@synthesize statusItemNormalTextEnabled;
@synthesize statusItemItsYourTurnTextEnabled;
@synthesize notifyAnimateStatusItemIntervalEnabled;
@synthesize minimumValueMonitoringIntervalAmount;


// -----------------------------------------------------------------------------
/// @brief Initializes an DGSMonXPreferences object.
///
/// @note This is the designated initializer of DGSMonXPreferences.
// -----------------------------------------------------------------------------
- (id) init
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;

  [self updateEnablersDependingOnStatusItemStyle];
  [self updateEnablersDependingOnNotifyStyle];
  [self updateMinimumValueMonitoringIntervalAmount];
  [self adjustMonitoringIntervalAmountIfExceedsMinimumValue];

  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXPreferences object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  // These release() messages balance the retain() messages that were sent by
  // Cocoa's nib-loading code when it loaded MainMenu.xib.
  [m_preferencesDialog release];   // widgets in this dialog retain m_serversController
  [m_serversController release];   // retains the application model!

  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Registers application defaults in the "Registration" domain of
/// NSUserDefaults.
///
/// This class method must be called as early as possible, before any objects
/// of the application start accessing the user defaults.
///
/// @note dgsmonX does not support "resettable user defaults" in the preferences
/// dialog, therefore this method does not provide the NSUserDefaultsController
/// with initial values.
// -----------------------------------------------------------------------------
+ (void) setupDefaults
{
  // Make paranoid checks about the user defaults file to provide detailed
  // error handling. If any check fails, an exception with an appropriate
  // error message is thrown. The exception is caught and the error handler
  // is notified using the error message inside the exception.
  // Note: this mechanism also catches unexpected exceptions...
  @try
  {
    NSBundle* mainBundle = [NSBundle mainBundle];
    if (! mainBundle)
    {
      NSString* reason = @"Unable to access application bundle contents.";
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    NSString* defaultsPathName = [mainBundle pathForResource:registrationDomainDefaultsFileName ofType:nil];
    if (nil == defaultsPathName)
    {
      NSString* reason = [[[NSString stringWithString:@"The defaults file\n\n"]
        stringByAppendingString:registrationDomainDefaultsFileName]
        stringByAppendingString:@"\n\ncould not be found in the resources of the application bundle."];
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    NSFileManager* fileManager = [NSFileManager defaultManager];
    BOOL isDirectory;  // use BOOL instead of bool due to method signature
    if (! [fileManager fileExistsAtPath:defaultsPathName isDirectory:&isDirectory])
    {
      NSString* reason = [[[NSString stringWithString:@"The defaults file\n\n"]
        stringByAppendingString:defaultsPathName]
        stringByAppendingString:@"\n\ndoes not exist."];
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    if (! [fileManager isReadableFileAtPath:defaultsPathName])
    {
      NSString* reason = [[NSString stringWithString:@"No read access to defaults file\n\n"]
        stringByAppendingString:defaultsPathName];
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    NSDictionary* defaultsDictionary = [NSDictionary dictionaryWithContentsOfFile:defaultsPathName];
    if (! defaultsDictionary)
    {
      NSString* reason = [[[NSString stringWithString:@"Defaults file\n\n"]
        stringByAppendingString:defaultsPathName]
        stringByAppendingString:@"\n\ndoes not contain an NSDictionary."];
      @throw [NSException exceptionWithName:@"DGSMonXException" reason:reason userInfo:nil];
    }

    // Yippie, we did it...
    [[NSUserDefaults standardUserDefaults] registerDefaults:defaultsDictionary];

    // Upgrade user defaults data before others access it. For this to
    // work we need to have the registration defaults in place.
    [DGSMonXUserDefaultsUpdater upgrade];
  }
  @catch(NSException* exception)
  {
    // Notify the error handler that an error has occurred
    [[NSNotificationCenter defaultCenter] postNotificationName:errorConditionOccurredNotification
                                                        object:[exception reason]];
  }
}

// -----------------------------------------------------------------------------
/// @brief Opens the user preferences dialog and displays it to the user.
///
/// @note The GUI controls in the dialog are "automatically" filled with values
/// by way of the "Cocoa Bindings" mechanism.
// -----------------------------------------------------------------------------
- (void) showPreferencesDialog
{
  // Make dgsmonX the active app so that the window we are going to display
  // is guaranteed to appear in front of other applications' windows.
  // Note: This is only necessary because dgsmonX is faceless and cannot become
  // the active application through normal user interaction.
  NSApplication* theApp = [NSApplication sharedApplication];
  [theApp activateIgnoringOtherApps:YES];

  // Prepare and run the dialog. The sheet is not attached to any window, it
  // is simply run as a modal window
  [theApp beginSheet:m_preferencesDialog modalForWindow:nil modalDelegate:nil didEndSelector:nil contextInfo:nil];
  [theApp runModalForWindow:m_preferencesDialog];

  // -------------------------------------------------
  // --- At this point the modal dialog is running ---
  // -------------------------------------------------

  // Clean up and close the dialog
  [theApp endSheet:m_preferencesDialog];
  // Remove dialog from screen
  [m_preferencesDialog orderOut:self];
}

// -----------------------------------------------------------------------------
/// @brief Reacts to the user clicking the button labelled "Close" in the user
/// preferences dialog.
///
/// Simulates the user clicking the red "close" button, which leads to the
/// NSWindow delegate method windowWillClose:() being invoked.
// -----------------------------------------------------------------------------
- (void) preferencesDialogCloseClicked:(id)sender
{
  [m_preferencesDialog performClose:self];
}

// -----------------------------------------------------------------------------
/// @brief Reacts to the user clicking the red "close" button of the NSWindow
/// that is the preferences dialog.
///
/// Closes the dialog.
///
/// @note The changes that the user made while the dialog was open have been
/// written to NSUserDefaults "automatically" by way of the "Cocoa Bindings"
/// mechanism.
///
/// @note This is an NSWindow delegate method.
// -----------------------------------------------------------------------------
- (void)windowWillClose:(NSNotification*)notification
{
  // Invoking this method closes the modal dialog and returns control to
  // showPreferencesDialog().
  [[NSApplication sharedApplication] stopModal];
}

// -----------------------------------------------------------------------------
/// @brief Reacts to the user clicking the "test" button in the user
/// preferences dialog.
///
/// The currently selected server is tested if its URL can be found and if it
/// is possible to request a status for the current user.
// -----------------------------------------------------------------------------
- (void) testClicked:(id)sender
{
  if (! m_serversController)
    return;
  NSArray* selectedObjects = [m_serversController selectedObjects];
  if (! selectedObjects || [selectedObjects count] != 1)
    return;
  DGSMonXServer* selectedServer = (DGSMonXServer*)[selectedObjects objectAtIndex:0];
  if (! selectedServer)
    return;
  int connectionResult = [selectedServer testConnection];
  NSString* messageText;
  NSString* infoText;
  if (connectionResult >= 0)
  {
    messageText = @"Connection successful.";
    infoText = [NSString stringWithFormat:@"Number of running games found: %d", connectionResult];
  }
  else
  {
    messageText = @"Connection failed.";
    infoText = @"Reason: ";
    if (-1 == connectionResult)
      infoText = [infoText stringByAppendingString:@"URL not found"];
    else if (-2 == connectionResult)
      infoText = [infoText stringByAppendingString:@"URL found, but status request failed"];
    else
      infoText = [infoText stringByAppendingString:@"Unknown"];
  }
  NSAlert* alert = [NSAlert alertWithMessageText:messageText
                                   defaultButton:@"OK"
                                 alternateButton:nil
                                     otherButton:nil
                       informativeTextWithFormat:infoText];
  [alert setAlertStyle:NSInformationalAlertStyle];
  [alert runModal];
}

// -----------------------------------------------------------------------------
/// @brief KVC getter for user defaults property "StatusItemStyle".
///
/// Normally this would not be needed and Cocoa Bindings would get/set its value
/// directly from/to NSUserDefaults. However, there are control enablers whose
/// state depends on the value of this property. To be able to react to changes
/// to this property, we therefore need to implement the setter of this
/// property in DGSMonXPreferences. Since we cannot specify different
/// controllers for getter and setter, we need to implement the getter here
/// as well. The implementation simply delegates the access to NSUserDefaults.
// -----------------------------------------------------------------------------
- (enum DGSMonXStatusItemStyle) statusItemStyle
{
  return [[NSUserDefaults standardUserDefaults] integerForKey:statusItemStyleKey];
}

// -----------------------------------------------------------------------------
/// @brief KVC setter for user defaults property "StatusItemStyle".
///
/// See the statusItemStyle() method documentation for details why
/// DGSMonXPreferences must implement this accessor.
// -----------------------------------------------------------------------------
- (void) setStatusItemStyle:(enum DGSMonXStatusItemStyle)statusItemStyle
{
  [[NSUserDefaults standardUserDefaults] setInteger:statusItemStyle forKey:statusItemStyleKey];
  [self updateEnablersDependingOnStatusItemStyle];
}

// -----------------------------------------------------------------------------
/// @brief Updates the Enabled state of the controls that depend on the
/// value of the "StatusItemStyle" user defaults property.
// -----------------------------------------------------------------------------
- (void) updateEnablersDependingOnStatusItemStyle
{
  bool styleWithIcon = false;
  bool styleWithText = false;
  switch ([self statusItemStyle])
  {
    case StatusItemStyleIconOnly:
      styleWithIcon = true;
      styleWithText = false;
      break;
    case StatusItemStyleTextOnly:
      styleWithIcon = false;
      styleWithText = true;
      break;
    case StatusItemStyleBothIconAndText:
      styleWithIcon = true;
      styleWithText = true;
      break;
    default:
      break;
  }
  self.statusItemIconStyleEnabled = styleWithIcon;
  self.statusItemNormalTextEnabled = styleWithText;
  self.statusItemItsYourTurnTextEnabled = styleWithText;
}

// -----------------------------------------------------------------------------
/// @brief KVC getter for user defaults property "StatusItemNormalText".
///
/// Normally this would not be needed and Cocoa Bindings would get/set its value
/// directly from/to NSUserDefaults. However, the property also requires input
/// validation which is not provided by NSUserDefaults. Unfortunately we cannot
/// specify two different controllers in Interface Builder: DGSMonXPreferences
/// for input validation, and NSUserDefaults for normal accessors. We therefore
/// need to implement the normal accessors in DGSMonXPreferences as well. The
/// implementation simply delegates the access to NSUserDefaults.
// -----------------------------------------------------------------------------
- (NSString*) statusItemNormalText
{
  return [[NSUserDefaults standardUserDefaults] stringForKey:statusItemNormalTextKey];
}

// -----------------------------------------------------------------------------
/// @brief KVC setter for user defaults property "StatusItemNormalText".
///
/// See the statusItemNormalText() method documentation for details why
/// DGSMonXPreferences must implement this accessor.
// -----------------------------------------------------------------------------
- (void) setStatusItemNormalText:(NSString*)text
{
  [[NSUserDefaults standardUserDefaults] setValue:text forKey:statusItemNormalTextKey];
}

// -----------------------------------------------------------------------------
/// @brief KVC validator for user defaults property "StatusItemNormalText".
// -----------------------------------------------------------------------------
- (BOOL) validateStatusItemNormalText:(id*)valueObject error:(NSError**)outError
{
  return [self validateStatusItemText:valueObject error:outError];
}

// -----------------------------------------------------------------------------
/// @brief KVC getter for user defaults property "StatusItemItsYourTurnText".
///
/// See the statusItemNormalText() method documentation for details why
/// DGSMonXPreferences must implement this accessor.
// -----------------------------------------------------------------------------
- (NSString*) statusItemItsYourTurnText
{
  return [[NSUserDefaults standardUserDefaults] stringForKey:statusItemItsYourTurnTextKey];
}

// -----------------------------------------------------------------------------
/// @brief KVC setter for user defaults property "StatusItemItsYourTurnText".
///
/// See the statusItemNormalText() method documentation for details why
/// DGSMonXPreferences must implement this accessor.
// -----------------------------------------------------------------------------
- (void) setStatusItemItsYourTurnText:(NSString*)text
{
  [[NSUserDefaults standardUserDefaults] setValue:text forKey:statusItemItsYourTurnTextKey];
}

// -----------------------------------------------------------------------------
/// @brief KVC validator for user defaults property "StatusItemItsYourTurnText".
// -----------------------------------------------------------------------------
- (BOOL) validateStatusItemItsYourTurnText:(id*)valueObject error:(NSError**)outError
{
  return [self validateStatusItemText:valueObject error:outError];
}

// -----------------------------------------------------------------------------
/// @brief Common backend for KVC validators for user defaults properties
/// "StatusItemNormalText" and "StatusItemItsYourTurnText".
// -----------------------------------------------------------------------------
- (BOOL) validateStatusItemText:(id*)valueObject error:(NSError**)outError
{
  NSString* text = (NSString*)*valueObject;
  if (nil == text || 0 == text.length)
  {
    if (outError != NULL)
    {
      NSString* errorString = @"Status item text must not be empty. Please enter at least one character.";
      NSDictionary* userInfoDict = [NSDictionary dictionaryWithObject:errorString
                                                               forKey:NSLocalizedDescriptionKey];
      *outError = [NSError errorWithDomain:errorDomain code:ErrorCodeStatusItemTextIsEmpty userInfo:userInfoDict];
    }
    return NO;
  }
  if (0 == [text stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]].length)
  {
    if (outError != NULL)
    {
      NSString* errorString = @"Status item text must not consist of only whitespace. Please enter at least one non-whitespace character.";
      NSDictionary* userInfoDict = [NSDictionary dictionaryWithObject:errorString
                                                               forKey:NSLocalizedDescriptionKey];
      *outError = [NSError errorWithDomain:errorDomain code:ErrorCodeStatusItemTextIsWhitespace userInfo:userInfoDict];
    }
    return NO;
  }
  return YES;
}

// -----------------------------------------------------------------------------
/// @brief KVC getter for user defaults property "NotifyStyle".
///
/// See the statusItemStyle() method documentation for details why
/// DGSMonXPreferences must implement this accessor.
// -----------------------------------------------------------------------------
- (enum DGSMonXNotificationStyle) notifyStyle
{
  return [[NSUserDefaults standardUserDefaults] integerForKey:notifyStyleKey];
}

// -----------------------------------------------------------------------------
/// @brief KVC setter for user defaults property "NotifyStyle".
///
/// See the statusItemStyle() method documentation for details why
/// DGSMonXPreferences must implement this accessor.
// -----------------------------------------------------------------------------
- (void) setNotifyStyle:(enum DGSMonXNotificationStyle)notificationStyle
{
  [[NSUserDefaults standardUserDefaults] setInteger:notificationStyle forKey:notifyStyleKey];
  [self updateEnablersDependingOnNotifyStyle];
}

// -----------------------------------------------------------------------------
/// @brief Updates the Enabled state of the controls that depend on the
/// value of the "NotifyStyle" user defaults property.
// -----------------------------------------------------------------------------
- (void) updateEnablersDependingOnNotifyStyle
{
  switch ([self notifyStyle])
  {
    case NotificationStyleAnimation:
      self.notifyAnimateStatusItemIntervalEnabled = true;
      break;
    default:
      self.notifyAnimateStatusItemIntervalEnabled = false;
      break;
  }
}

// -----------------------------------------------------------------------------
/// @brief KVC getter for user defaults property "MonitoringIntervalUnit".
///
/// Normally this would not be needed and Cocoa Bindings would get/set its value
/// directly from/to NSUserDefaults. However, changing the value of this
/// property has an influence on the minimum value of another property
/// "MonitoringIntervalAmount". If that other property's current value is
/// below its new minimum value, the current value even needs to be adjusted.
///
/// To be able to react to changes to "MonitoringIntervalUnit", we would need
/// to implement only the setter, but since Cocoa Bindings does not allow us
/// to specify different controllers for getter and setter, we need to implement
/// the getter here as well. The implementation simply delegates the access to
/// NSUserDefaults.
// -----------------------------------------------------------------------------
- (int) monitoringIntervalUnit
{
  return [[NSUserDefaults standardUserDefaults] integerForKey:monitoringIntervalUnitKey];
}

// -----------------------------------------------------------------------------
/// @brief KVC setter for user defaults property "MonitoringIntervalUnit".
///
/// See the monitoringIntervalUnit() method documentation for details why
/// DGSMonXPreferences must implement this accessor.
// -----------------------------------------------------------------------------
- (void) setMonitoringIntervalUnit:(int)monitoringIntervalUnit
{
  [[NSUserDefaults standardUserDefaults] setInteger:monitoringIntervalUnit forKey:monitoringIntervalUnitKey];

  [self updateMinimumValueMonitoringIntervalAmount];
  [self adjustMonitoringIntervalAmountIfExceedsMinimumValue];
}

// -----------------------------------------------------------------------------
/// @brief Updates the minimum value of the user defaults property
/// "MonitoringIntervalAmount". The new minimum value depends on the current
/// value of another user defaults property "MonitoringIntervalUnit".
// -----------------------------------------------------------------------------
- (void) updateMinimumValueMonitoringIntervalAmount
{
  switch ([self monitoringIntervalUnit])
  {
    case SecondsTimeUnit:
      self.minimumValueMonitoringIntervalAmount = 300;
      break;
    case MinutesTimeUnit:
      self.minimumValueMonitoringIntervalAmount = 5;
      break;
    default:
      self.minimumValueMonitoringIntervalAmount = 1;
      break;
  }
}

// -----------------------------------------------------------------------------
/// @brief Adjusts the current value of the user defaults property
/// "MonitoringIntervalAmount" if it exceeds the allowed minimum value.
// -----------------------------------------------------------------------------
- (void) adjustMonitoringIntervalAmountIfExceedsMinimumValue
{
  int monitoringIntervalAmount = [[NSUserDefaults standardUserDefaults] integerForKey:monitoringIntervalAmountKey];
  if (monitoringIntervalAmount < self.minimumValueMonitoringIntervalAmount)
    [[NSUserDefaults standardUserDefaults] setInteger:minimumValueMonitoringIntervalAmount forKey:monitoringIntervalAmountKey];
}

@end
