// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXThread.h"
#import "DGSMonXController.h"
#import "DGSMonXServer.h"


/// @brief This category declares private methods for the DGSMonXThread
/// class. 
@interface DGSMonXThread()
/// @name Initialization + deallocation
//@{
- (id) init;
- (id) initWithPorts:(NSArray*)portArray;
- (void) dealloc;
//@}

/// @name DGSMonXServerCheckDelegate protocol method
//@{
- (void) gameStateCheckDidFinish:(DGSMonXServerCheck*)serverCheck;
//@}

/// @name Internal helpers
//@{
- (NSTimeInterval) getCheckInterval;
- (void) runAutomaticChecks;
- (void) runManualCheck:(id)server;
- (bool) shouldExit;
- (void) prepareToDie;
- (NSTimer*) scheduleAutomaticChecks;
//@}
@end


@implementation DGSMonXThread

// -----------------------------------------------------------------------------
/// @brief This is the main function of a thread.
///
/// This method creates a DGSMonXThread instance, announces that instance back
/// to the main thread and repeatedly runs the thread's run loop, until the
/// DGSMonXThread instance returns true for its shouldExit() method. At this
/// time, the DGSMonXThread instance is deallocated and the thread terminates.
// -----------------------------------------------------------------------------
+ (void)main:(NSArray*)portArray
{
  // Create an autorelease pool as the very first thing in this thread
  NSAutoreleasePool* mainPool = [[NSAutoreleasePool alloc] init];

  // Create the thread object. It will callback to the main thread to announce
  // its instance. It will also schedule a game state check that will be
  // executed immediately when we start the run loop below.
  DGSMonXThread* threadObject = [[self alloc] initWithPorts:portArray];

  // The following control loop re-runs the thread's run loop until threadObject
  // tells it to stop (shouldExit() returns true). The control loop gets a
  // chance to query threadObject each time the runMode:beforeDate:() method
  // returns. This happens
  // - When "distantFuture" is reached - which essentially means "never"
  // - Or after every input event that the run loop processes. Input events
  //   occur when the main thread sends messages through the ports in portArray.
  //   The message is processed first, then the run loop exits (i.e.
  //   runMode:beforeDate:() returns). For instance, the main thread tells us
  //   to suspend monitoring when the preferences dialog is opened. The
  //   all-important input event that causes our own control loop to break is
  //   the main thread's exit() message, which will tell threadObject to return
  //   true for shouldExit().
  // - It does ***NOT*** happen when a timer fires.
  NSDate* distantFuture = [NSDate distantFuture]; 
  while (true)
  {
    // Create a new pool & drain it in each iteration, otherwise objects will
    // accumulate in the outer main pool but will never be released until
    // the application is terminated
    NSAutoreleasePool* loopPool = [[NSAutoreleasePool alloc] init];
    // We have no reason to use something else than NSDefaultRunLoopMode, but we
    // have an excellent reason why we *must* use that mode: Game state checks
    // are performed using NSURLConnection, and this class requires that the
    // run loop be started in default mode!
    bool hasInputSources = [[NSRunLoop currentRunLoop] runMode:NSDefaultRunLoopMode
                                                    beforeDate:distantFuture];
    // Drain the pool, thus releasing objects that have accumulated during the
    // last iteration
    [loopPool drain];
    // hasInputSources should always be true, because the run loop should
    // always have input sources (the NSPort objects in portArray).
    if (! hasInputSources)
      break;
    if ([threadObject shouldExit])
      break;
  }

  // Deallocate the thread object
  [threadObject prepareToDie];
  [threadObject release];

  // Deallocate the autorelease pool as the very last thing in this thread
  [mainPool release];
}

// -----------------------------------------------------------------------------
/// @brief This initializer always returns @e nil.
// -----------------------------------------------------------------------------
- (id) init
{
  // Invoke designated initializer, which will always return nil
  return [self initWithPorts:nil];
}

// -----------------------------------------------------------------------------
/// @brief Initializes a DGSMonXThread object.
///
/// @a portArray is expected to contain 2 NSPort objects that can be used to
/// create an NSConnection for communicating with the application main thread.
///
/// @note This is the designated initializer of DGSMonXThread.
// -----------------------------------------------------------------------------
- (id) initWithPorts:(NSArray*)portArray
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;

  // Create connection for communication with main thread
  m_connection = [[NSConnection alloc] initWithReceivePort:[portArray objectAtIndex:0]
                                                  sendPort:[portArray objectAtIndex:1]];
  // Retains self, creating a retain cycle. The cycle is broken when the
  // connection's invalidate() method is invoked - see prepareToDie().
  [m_connection setRootObject:self];

  // Callback to the main thread and announce this thread instance
  [(DGSMonXController*)[m_connection rootProxy] setMonitoringThread:self];

  // Initialize other members
  m_suspendMonitoring = false;
  m_shouldExit = false;
  m_automaticChecksTimer = nil;
  m_immediatelyRunChecks = false;
  m_pendingCheckResults = [[NSMutableArray array] retain];

  // Schedule a game state check that is run immediately when the thread enters
  // its run loop for the first time
  m_automaticChecksTimer = [NSTimer scheduledTimerWithTimeInterval:0
                                                            target:self
                                                          selector:@selector(runAutomaticChecks)
                                                          userInfo:nil
                                                           repeats:NO];

  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXThread object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  [m_connection autorelease];
  [m_pendingCheckResults autorelease];

  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Calculates the interval in seconds between game state checks.
/// Calculation is done using the current user defaults.
///
/// Returns -1 if any error occurs.
// -----------------------------------------------------------------------------
- (NSTimeInterval) getCheckInterval
{
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  if (! userDefaults)
    return -1;

  int unit = [userDefaults integerForKey:monitoringIntervalUnitKey];
  int amount = [userDefaults integerForKey:monitoringIntervalAmountKey];

  if (HoursTimeUnit == unit)
    return 3600 * amount;
  else if (MinutesTimeUnit == unit)
    return 60 * amount;
  else if (SecondsTimeUnit == unit)
    return amount;

  return -1;
}

// -----------------------------------------------------------------------------
/// @brief Triggers a series of game state checks for all servers currently
/// available from the application model for which monitoring is enabled.
///
/// The checks are performed asynchronously, i.e. they are going to happen only
/// after this method returns. The reason for this is that NSURLConnection
/// (which is used for implementing the game state checks) requires an active
/// run loop in the thread that uses it. The run loop of DGSMonXThread becomes
/// active again only after this method returns.
///
/// After all checks have been triggered, the next series of checks is
/// scheduled according to the current checking interval.
///
/// This method is intended to run checks that have been automatically
/// scheduled. Read the class documentation for details.
///
/// @note This method must be invoked only by one source: By the timer
/// referenced in #m_automaticChecksTimer.
// -----------------------------------------------------------------------------
- (void) runAutomaticChecks
{
  // Because this method can be called only by this timer, and because we know
  // that the timer is single-shot, we know that the timer has expired and will
  // soon be released
  // -> whatever happens during the remainder of this function, we must remove
  //    the reference to the NSTimer object
  m_automaticChecksTimer = nil;

  // Abort if monitoring is currently suspended. Remember to re-run checks
  // as soon as monitoring is resumed.
  if (m_suspendMonitoring)
  {
    m_immediatelyRunChecks = true;
    return;
  }
  m_immediatelyRunChecks = false;

  // Iterate & process items
  // Note: Servers are checked in the order in which they appear in the model.
  NSArray* serverList = [(DGSMonXController*)[m_connection rootProxy] serverList];
  NSEnumerator* enumerator = [serverList objectEnumerator];
  bool threadState = IdleThreadState;
  while (enumerator)
  {
    DGSMonXServer* server = (DGSMonXServer*)[enumerator nextObject];
    if (! server)
      break;
    if (! [server monitored])
      continue;
    // Creating an DGSMonXServerCheck object immediately starts the game state
    // check. The initializing method returns immediately, though, and the check
    // is performed asynchronously later on, as soon the run loop of this
    // thread becomes active again.
    DGSMonXServerCheck* serverCheck = [DGSMonXServerCheck serverCheckWithServer:server delegate:self];
    // NSMutableArray retains the DGSMonXServerCheck object for us
    [m_pendingCheckResults addObject:serverCheck];
    // It is now safe to set the thread state to "checking"
    // -> because we have at least one server that is being checked, we can be
    //    sure that gameStateCheckDidFinish:() will be called at least once
    // -> gameStateCheckDidFinish:() will reset the thread state to "idle"
    threadState = CheckingThreadState;
  }

  // Tell the main thread what we are doing
  [(DGSMonXController*)[m_connection rootProxy] setMonitoringThreadState:threadState];

  // Reschedule the next series of automatic checks
  m_automaticChecksTimer = [self scheduleAutomaticChecks];
}

// -----------------------------------------------------------------------------
/// @brief Triggers a single game state check for server @a server.
///
/// The check is performed asynchronously. See runAutomaticChecks() for details
/// about this.
///
/// This method is intended to run checks that have been manually scheduled.
/// Read the class documentation for details.
// -----------------------------------------------------------------------------
- (void) runManualCheck:(id)server
{
  // Discard scheduled check, no re-scheduling for manual checks!
  if (m_suspendMonitoring)
    return;
  // Tell the main thread what we are doing
  [(DGSMonXController*)[m_connection rootProxy] setMonitoringThreadState:CheckingThreadState];
  // The check starts immediately, but then proceeds on asynchronously
  DGSMonXServerCheck* serverCheck = [DGSMonXServerCheck serverCheckWithServer:server delegate:self];
  // NSMutableArray retains the DGSMonXServerCheck object for us
  [m_pendingCheckResults addObject:serverCheck];
}

// -----------------------------------------------------------------------------
/// @brief Schedules a series of game state checks for all servers, then returns
/// immediately. The checks are run as soon as possible.
///
/// The game state check for each server is scheduled separately by invoking
/// checkSingle:(). As a consequence of this, no check is scheduled for a server
/// if one is already in progress, or if monitoring is disabled for the server.
// -----------------------------------------------------------------------------
- (void) checkAll
{
  NSArray* serverList = [(DGSMonXController*)[m_connection rootProxy] serverList];
  NSEnumerator* enumerator = [serverList objectEnumerator];
  while (enumerator)
  {
    DGSMonXServer* server = (DGSMonXServer*)[enumerator nextObject];
    if (! server)
      break;
    // Delegate all checks
    [self checkSingle:server];
  }
}

// -----------------------------------------------------------------------------
/// @brief Schedules a single game state check for server @a server, then
/// returns immediately. The check is run as soon as possible.
///
/// Calling this method has no effect if a game state check for @a server is
/// already in progress, or if monitoring is disabled for @a server.
// -----------------------------------------------------------------------------
- (void) checkSingle:(DGSMonXServer*)server
{
  if (! [server monitored])
    return;
  NSEnumerator* enumerator = [m_pendingCheckResults objectEnumerator];
  while (enumerator)
  {
    DGSMonXServerCheck* serverCheck = (DGSMonXServerCheck*)[enumerator nextObject];
    if (! serverCheck)
      break;
    if ([[serverCheck server] uniqueID] == [server uniqueID])
      return;
  }
  // Using this NSObject method allows us to specify the server object as a
  // direct argument to the target method. With NSTimer we would have to put the
  // argument into a dictionary, and then extract it again in the target
  // method. That would be too much trouble for our simple needs.
  [self performSelector:@selector(runManualCheck:)
             withObject:server
             afterDelay:0];
}

// -----------------------------------------------------------------------------
/// @brief Cancels all game state checks that are still in progress.
///
/// Calling this method has no effect if no game state checks are in progress.
///
/// @note Results for those game state checks that have already finished remain
/// intact.
// -----------------------------------------------------------------------------
- (void) cancelAll
{
  NSEnumerator* enumerator = [m_pendingCheckResults objectEnumerator];
  while (enumerator)
  {
    DGSMonXServerCheck* serverCheck = (DGSMonXServerCheck*)[enumerator nextObject];
    if (! serverCheck)
      break;
    [serverCheck cancel];
  }
  [m_pendingCheckResults removeAllObjects];
  [(DGSMonXController*)[m_connection rootProxy] gameStateCheckDidFinish];
  [(DGSMonXController*)[m_connection rootProxy] setMonitoringThreadState:IdleThreadState];
}

// -----------------------------------------------------------------------------
/// @brief Cancels a single game state check that is still in progress for
/// server @a server.
///
/// Calling this method has no effect if no game state check is in progress
/// for @a server.
// -----------------------------------------------------------------------------
- (void) cancelSingle:(DGSMonXServer*)server
{
  NSEnumerator* enumerator = [m_pendingCheckResults objectEnumerator];
  while (enumerator)
  {
    DGSMonXServerCheck* serverCheck = (DGSMonXServerCheck*)[enumerator nextObject];
    if (! serverCheck)
      break;
    if ([[serverCheck server] uniqueID] == [server uniqueID])
    {
      [serverCheck cancel];
      [m_pendingCheckResults removeObject:serverCheck];
      [(DGSMonXController*)[m_connection rootProxy] gameStateCheckDidFinish];
      if ([m_pendingCheckResults count] == 0)
        [(DGSMonXController*)[m_connection rootProxy] setMonitoringThreadState:IdleThreadState];
      return;
    }
  }
}

// -----------------------------------------------------------------------------
/// @brief Suspends all future monitoring activities of this thread and cancels
/// game state checks that are still in progress.
// -----------------------------------------------------------------------------
- (void) suspendMonitoring
{
  m_suspendMonitoring = true;
  // If there are pending game state checks, cancel them. Also set
  // m_immediatelyRunChecks to true so that a new series of automatic checks is
  // run as soon as monitoring is resumed.
  // Note: This also happens if the pending checks are manual checks!
  if ([m_pendingCheckResults count] > 0)
  {
    [self cancelAll];
    m_immediatelyRunChecks = true;
  }
}

// -----------------------------------------------------------------------------
/// @brief Resumes monitoring activities of this thread.
///
/// Possibly schedules an immediate series of game state checks, e.g. if a
/// scheduled check was due while monitoring was suspended.
// -----------------------------------------------------------------------------
- (void) resumeMonitoring
{
  m_suspendMonitoring = false;
  if (m_immediatelyRunChecks)
  {
    m_automaticChecksTimer = [NSTimer scheduledTimerWithTimeInterval:0
                                                              target:self
                                                            selector:@selector(runAutomaticChecks)
                                                            userInfo:nil
                                                             repeats:NO];
  }
}

// -----------------------------------------------------------------------------
/// @brief Informs this thread that the monitoring interval has changed.
///
/// The next series of automatic checks that is scheduled is cancelled, and a
/// new series of checks using the new interval is scheduled instead.
///
/// Any checks running right now are left running. Any checks scheduled for
/// immediatey execution (e.g. from an invocation of checkAll()) are left
/// scheduled.
///
/// @note Although it is not limited to a specific event or time during the
/// application's runtime, this method is typically invoked after the user has
/// made changes in the Preferences dialog.
// -----------------------------------------------------------------------------
- (void) monitoringIntervalHasChanged
{
  if (! m_automaticChecksTimer)
    return;
  // Timer will be released by the run loop - at least that's what we hope...
  [m_automaticChecksTimer invalidate];
  m_automaticChecksTimer = [self scheduleAutomaticChecks];
}

// -----------------------------------------------------------------------------
/// @brief Returns true if this thread should terminate.
// -----------------------------------------------------------------------------
- (bool) shouldExit;
{
  return m_shouldExit;
}

// -----------------------------------------------------------------------------
/// @brief Sets the termination flag of this thread to true (i.e. shouldExit()
/// will return true after this method has been invoked).
// -----------------------------------------------------------------------------
- (void) exit
{
  m_shouldExit = true;
}

// -----------------------------------------------------------------------------
/// @brief Sent when @a serverCheck has finished its game state check.
///
/// When all pending checks have been performed, the main thread is notified so
/// that it can update the GUI.
///
/// @note This is an DGSMonXServerCheckDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) gameStateCheckDidFinish:(DGSMonXServerCheck*)serverCheck
{
  // NSMutableArray releases the DGSMonXServerCheck object for us when we
  // remove it.
  [m_pendingCheckResults removeObject:serverCheck];
  if ([m_pendingCheckResults count] == 0)
  {
    [(DGSMonXController*)[m_connection rootProxy] gameStateCheckDidFinish];
    [(DGSMonXController*)[m_connection rootProxy] setMonitoringThreadState:IdleThreadState];
  }
}

// -----------------------------------------------------------------------------
/// @brief Sent by the monitoring thread's main:() method just before it is
/// releasing this DGSMonXThread object.
///
/// This method must make sure that the release will lead to deallocation of
/// this DGSMonXServer object.
// -----------------------------------------------------------------------------
- (void) prepareToDie
{
  // Invalidates all kinds of things - see the method docs for details. The
  // important thing for us is that the invalidating process causes the
  // connection to release the retained reference to ourself (i.e. the
  // DGSMonXThread instance). This breaks the retain cycle that we created in
  // our initWithPorts:() method.
  [m_connection invalidate];
}

// -----------------------------------------------------------------------------
/// @brief Schedules a series of automatic checks and returns the NSTimer
/// object that was used for scheduling.
///
/// Uses the monitoring interval returned by getCheckInterval() to schedule the
/// series of checks.
///
/// When the timer fires, it will invoke runAutomaticChecks().
// -----------------------------------------------------------------------------
- (NSTimer*) scheduleAutomaticChecks
{
  // Re-schedule the next series of checks
  // - Timer is retained by NSRunLoop
  // - Timer retains this thread object (self)
  NSTimeInterval checkInterval = [self getCheckInterval];
  if (checkInterval < 0)
  {
    [[NSNotificationCenter defaultCenter] postNotificationName:errorConditionOccurredNotification
                                                        object:@"Could not calculate timer interval from user defaults."];
    // Recover from the error by using a hard-coded default value
    checkInterval = 3600;
  }
  return [NSTimer scheduledTimerWithTimeInterval:checkInterval
                                          target:self
                                        selector:@selector(runAutomaticChecks)
                                        userInfo:nil
                                         repeats:NO];
}

@end
